""" Modul pro vytváření sémantických embeddingů a dalších funkcí.
"""

import numpy as np
from sentence_transformers import SentenceTransformer

import nltk
nltk.download("punkt")

from nltk import tokenize


def get_embeddings(text: str, model: SentenceTransformer, sent_tokenize: bool = True) -> np.ndarray:
    """Převede text na sémantické embeddingy (vektor).

    Umožňuje převést kratší i delší text na embeddingy. Pokud je sent_tokenize True, tak nejdříve zákoduje jednotlivé \
    věty a pak jednotlivé věty zprůměruje k zákodování celého textu. V případě, že je False, tak zákoduje text jako \
    celek. Druhá varianta lépe zachytí celkový kontext, ale většina SentenceTransformer modelů umí najednou zpracovat \
    pouze určité množství tokenů (např. 128 tokenů, což to odpovídá cca třem průměrným větám), po tomto množství \
    kvalita klesá.

    :param text: text, který chceme zakódvat do embeddingů
    :param model: SentenceTransformer model, podle kterého se text zakóduje do embeddingů
    :param sent_tokenize: pokud je nastavený, tak nejdříve zakóduje jednotlivé věty a až pak celý text.
    :return: embeddingový vektor vstupního textu
    """

    if sent_tokenize:
        sentences = tokenize.sent_tokenize(text)
        sentence_embeddings = [model.encode(sent) for sent in sentences]
        doc_embedding = np.mean(sentence_embeddings, axis=0)
    else:
        doc_embedding = model.encode(text)

    return doc_embedding



"""
Tento modul slouží k načtení různých zdrojů dat. V tuto chvíli umožňuje načíst \
data o projektech a uchazečích/příjemcích ze "zdroje pravdy" a IS VaVaI.
"""

import gspread
import pandas as pd


class Projects:
    """ Třída, která načítá a reprezentuje tabulku projektů.

    Funguje pouze v rámci Google Colab prostředí.

    :param projects: DataFrame načtených dat ze zdroje nebo z nově vytvořené (vyfiltrované) instance
    :type projects: DataFrame
    :param summary: DataFrame s agregovanými údaji na úrovni veřejných soutěží
    :type summary: DataFrame
    """

    def __init__(self, creds_or_df: object):
        """ Kontstruktor, který načte data do DataFrame, očistí finanční hodnoty a vytvoří agregovanou tabulku.

        :param creds_or_df: údaje, které slouží k authenizaci v rámci Google Colab nebo přímo DataFrame projektu
        """
        self.projects = None
        if isinstance(creds_or_df, pd.DataFrame):
            self.projects = creds_or_df
        else:
            self.projects = self._get_projects(creds_or_df)
            self._finance_cleaning('Náklady celkem')
            self._finance_cleaning('Podpora celkem')
            self._finance_cleaning('Ostatní celkem')
        self.summary = self.create_summary()

    def _get_projects(self, creds_or_df: object) -> pd.DataFrame:
        """ Načte data o projektech ze "zdroje pravdy" z googlesheets uloženého na Google disku.

        Lze použít pouze v rámci Google Colab prostředí.

        :param creds_or_df: údaje, které slouží k authenizaci v rámci Google Colab
        :return: DataFrame načtených dat ze zdroje
        """
        file_id = "1Ax1OYkdg3IA1YZki0fePizgQR6zOuzq7VGhFgeMorDQ"

        gc = gspread.authorize(creds_or_df)
        sht = gc.open_by_key(file_id)
        worksheet = sht.get_worksheet(0)

        rows = worksheet.get_all_values()

        df = pd.DataFrame.from_records(rows[1:], columns=rows[0])

        return df

    def _finance_cleaning(self, column_name: str):
        """ Interní funkce, která očistí finanční data.

        :param column_name: název sloupce, ve kterém se mají finanční data očistit
        """

        self.projects[column_name].fillna(0, inplace=True)
        self.projects[column_name] = self.projects[column_name].str.replace(',', '.')
        self.projects[column_name] = self.projects[column_name].replace('', '0')
        self.projects[column_name] = self.projects[column_name].astype(float)

    def create_summary(self, level: str = 'cfp') -> pd.DataFrame:
        """ Vytvoří agregovaný souhrn buď na úrovni veřejných soutěží (defaultní) nebo na úrovni programů.

        :param level: určuje, na jaké úrovni se provede agregace

                      * 'cfp' (defaultní) - na úrovni veřejných soutěží
                      * 'prog' - na úrovni programů
        :return: agregovaný DataFrame, který obsahuje:

                * Počet podaných projektů
                * Počet podpořených projektů
                * Náklady podpořených projektů
                * Podpora podpořených projektů
        """
        if level not in ['cfp', 'prog']:
            raise ValueError('Neexistující forma agregace.')

        temp_df = self.projects.copy()
        temp_df['Podpořené'] = temp_df.apply(
            lambda x: 'Ano' if x['Fáze projektu'] in ['Realizace', 'Implementace', 'Ukončené'] else 'Ne', axis=1)
        submitted = temp_df.groupby(['Kód programu', 'Kód VS']).agg(
            {'Kód projektu': 'count', 'Náklady celkem': 'sum', 'Podpora celkem': 'sum'}).reset_index()
        funded = temp_df[temp_df['Podpořené'] == 'Ano'].groupby(['Kód programu', 'Kód VS']).agg(
            {'Kód projektu': 'count', 'Náklady celkem': 'sum', 'Podpora celkem': 'sum'}).reset_index()

        summary_df = pd.merge(submitted[['Kód programu', 'Kód VS', 'Kód projektu']], funded, how='inner',
                              on=['Kód programu', 'Kód VS'])
        summary_df.columns = ['Kód programu', 'Kód VS', 'Podané', 'Podpořené', 'Náklady', 'Podpora']

        if level == 'cfp':
            pass
        elif level == 'prog':
            summary_df = summary_df.groupby('Kód programu').agg('sum', numeric_only=True).reset_index()

        return summary_df

    def select_programme(self, *args: str) -> 'Projects':
        """ Vyfiltruje tabulku tak, aby obsahovala pouze projekty vybraných programů.

        :param args: kódy programů, které se mají vyfiltrovat
        :return: nová instance třídy Projects s vyfiltrovanými údaji
        :raise: ValueError
        """

        existing_programmes = self.projects['Kód programu'].unique()

        missing_programmes = [prog for prog in args if prog not in existing_programmes]

        if missing_programmes:
            raise ValueError(f'Programy {missing_programmes} neexistují.')

        else:
            programme_list = [prog for prog in args]
            select_df = self.projects[self.projects['Kód programu'].isin(programme_list)].reset_index(drop=True)
            return Projects(select_df)  # todo maybe another class Programms?

    def select_cfp(self, *args: str) -> 'Projects':
        """ Vyfiltruje tabulku tak, aby obsahovala pouze projekty vybraných veřejných soutěží.

        :param args: kódy veřejných soutěží, které se mají vyfiltrovat
        :return: nová instance třídy Projects s vyfiltrovanými údaji
        :raise: ValueError
        """
        existing_cfp = self.projects['Kód VS'].unique()

        missing_cfp = [cfp for cfp in args if cfp not in existing_cfp]

        if missing_cfp:
            raise ValueError(f'Veřejné soutěže {missing_cfp} neexistují.')

        else:
            cfp_list = [cfp for cfp in args]
            select_df = self.projects[self.projects['Kód VS'].isin(cfp_list)].reset_index(drop=True)
            return Projects(select_df)

    def select_funded(self) -> 'Projects':
        """ Vyfiltruje tabulku tak, aby obsahovala pouze podpořené projekty.

        :return: nová instance třídy Projects s vyfiltrovanými údaji
        """
        funded_states = ['Realizace', 'Implementace', 'Ukončené']
        select_df = self.projects[self.projects['Fáze projektu'].isin(funded_states)].reset_index(drop=True)
        return Projects(select_df)


def projects_finance(creds_or_df: object) -> pd.DataFrame:
    """ Načte data o financích projektů ze "zdroje pravdy" z googlesheets uloženého na Google disku.

    Finance jsou v rozdělení po jednotlivých letech.
    Lze použít pouze v rámci Google Colab prostředí.

    :param creds_or_df: údaje, které slouží k authenizaci v rámci Google Colab
    :return: DataFrame načtených dat ze zdroje
    """
    file_id = "1Ax1OYkdg3IA1YZki0fePizgQR6zOuzq7VGhFgeMorDQ"

    gc = gspread.authorize(creds_or_df)
    sht = gc.open_by_key(file_id)
    worksheet = sht.get_worksheet(1)

    rows = worksheet.get_all_values()

    df = pd.DataFrame.from_records(rows[1:], columns=rows[0])
    return df


def organizations(creds_or_df: object) -> pd.DataFrame:
    """ Načte data o uchazečích/příjemcích ze "zdroje pravdy" z googlesheets uloženého na Google disku.

    Lze použít pouze v rámci Google Colab prostředí.

    :param creds_or_df: údaje, které slouží k authenizaci v rámci Google Colab
    :return: DataFrame načtených dat ze zdroje
    """
    file_id = "1h7HpPn-G0_XY2gb_sExAQDkzR1TswGUH_2FuHCWhbRg"

    gc = gspread.authorize(creds_or_df)
    sht = gc.open_by_key(file_id)
    worksheet = sht.get_worksheet(0)

    rows = worksheet.get_all_values()

    df = pd.DataFrame.from_records(rows[1:], columns=rows[0])
    return df


def organizations_finance(creds_or_df: object) -> pd.DataFrame:
    """ Načte data o financích uchazečů/příjemců ze "zdroje pravdy" z googlesheets uloženého na Google disku.

    Finance jsou v rozdělení po jednotlivých letech.
    Lze použít pouze v rámci Google Colab prostředí.

    :param creds_or_df: údaje, které slouží k authenizaci v rámci Google Colab
    :return: DataFrame načtených dat ze zdroje
    """
    file_id = "1h7HpPn-G0_XY2gb_sExAQDkzR1TswGUH_2FuHCWhbRg"

    gc = gspread.authorize(creds_or_df)
    sht = gc.open_by_key(file_id)
    worksheet = sht.get_worksheet(1)

    rows = worksheet.get_all_values()

    df = pd.DataFrame.from_records(rows[1:], columns=rows[0])
    return df


def isvav_projects() -> pd.DataFrame:
    """ Načte data o příjemcích z otevřených dat IS VaVaI.

    Data jsou aktualizovaná cca jednou za čtvrt roku.

    :return: DataFrame načtených dat ze zdroje
    """
    df = pd.read_csv("https://www.isvavai.cz/dokumenty/opendata/CEP-projekty.csv")
    return df


def isvav_organizations() -> pd.DataFrame:
    """ Načte data o projektech z otevřených dat IS VaVaI.

    Data jsou aktualizovaná cca jednou za čtvrt roku.

    :return: DataFrame načtených dat ze zdroje
    """
    df = pd.read_csv("https://www.isvavai.cz/dokumenty/opendata/CEP-ucastnici.csv")
    return df


def results(creds_or_df: object) -> pd.DataFrame:
    """ Načte data o výsledcích ze "zdroje pravdy" z googlesheets uloženého na Google disku.

    Lze použít pouze v rámci Google Colab prostředí.

    :param creds_or_df: údaje, které slouží k authenizaci v rámci Google Colab
    :return: DataFrame načtených dat ze zdroje
    """
    file_id = "1eSE6gB8bwuP6OVwVVhQojQLS6aPi_q8t7gK6VhPyiRw"

    gc = gspread.authorize(creds_or_df)
    sht = gc.open_by_key(file_id)
    worksheet = sht.get_worksheet(0)

    rows = worksheet.get_all_values()

    df = pd.DataFrame.from_records(rows[1:], columns=rows[0])
    return df



"""Soubor funkcí, které slouží k hromadnému vytváření podkladů k tvorbě a úpravě data lineage \
v DataHubu přes API."""


def _create_field_lineage(entity_urn: str, upstream_urn: str, field_dict: dict[str, str]) -> dict[str, any]:
    """ Vytvoří data lineage na úrovni dvou fieldů.

    :param entity_urn: URN datasetu, ve kterém je field, pro které tvoříme data lineage
    :param upstream_urn: URN datasetu, ve kterém je field, na který má vazbu (lineage)
    :param field_dict: informace a filed level lineage
    :return: dict s informacemi o vazbě mezi dvěma fieldy
    """

    field_upstream_name = field_dict['upstream_lineage_field']
    field_upstream_urn = f'urn:li:schemaField:({upstream_urn},{field_upstream_name})'
    field_upstream_type = field_dict['upstream_field_type']

    field_downstream_name = field_dict['downstream_lineage_field']
    field_downstream_urn = f'urn:li:schemaField:({entity_urn},{field_downstream_name})'
    field_downstream_type = field_dict['downstream_field_type']

    field_lineage = {
        'upstreamType': field_upstream_type,
        'upstreams': [
            field_upstream_urn
        ],
        'downstreamType': field_downstream_type,
        'downstreams': [
            field_downstream_urn
        ]
    }

    return field_lineage


def create_dataset_lineage(entity_name: str, platform: str, lineage_dict: dict[str, any]
                           , field_level: bool = False) -> dict[str, any]:
    """ Vytvoří data lineage (vazby) pro daný dataset.

    Umožňuje vytvářet data lineage na úrovni datasetů i na úrovni jednotlivých fieldů.

    :param entity_name: název datasetu, pro který se data lineage vytváří
    :param platform: název platformy, která je zdrojem datasetu
                     (např. etalon, googlesheets, ISTA, Postgres, OpenAPI...)
    :param lineage_dict: dict datasetu, který obsahuje jednotlivé vazby (na úrovni datasetů nebo na úrovni fieldů)
    :param field_level: informace, jestli obsahuje dataset lineage na úrovni fieldů
    :return: dict vazeb (data lineage) pro daný dataset
    """

    entity_urn = f'urn:li:dataset:(urn:li:dataPlatform:{platform},{entity_name},PROD)'

    upstreams = []
    field_lineages = []

    for key, value in lineage_dict.items():
        upstream_name = value['upstream_lineage_dataset']
        upstream_platform = value['upstream_platform']
        upstream_urn = f'urn:li:dataset:(urn:li:dataPlatform:{upstream_platform},{upstream_name},PROD)'
        upstream_type = value['upstream_dataset_type']

        upstream_dict = {
            'dataset': upstream_urn,
            'type': upstream_type
        }

        # pro field_lineage bude existovat více záznamů pro vazbu dvou datasetů, proto se "filtruje"
        if upstream_dict not in upstreams:
            upstreams.append(upstream_dict)

        if field_level:
            field_lineage = _create_field_lineage(entity_urn, upstream_urn, value)
            field_lineages.append(field_lineage)

    upstream_lineage = {
        "entityType": "dataset",
        "entityUrn": entity_urn,
        "aspect": {
            "__type": "UpstreamLineage",
            "upstreams": upstreams,
            "fineGrainedLineages": field_lineages
        }
      }

    return upstream_lineage

# TODO: rozdílné funkce na vytvoření nové data lineage a update stávající data lineage



from setuptools import setup, find_packages

setup(
    name='tacrpy',
    version='0.2.0',
    description='Analytická knihovna pro potřeby TA ČR',
    long_description='Knihovna, která slouží pro práci s daty a vypracování analýz TA ČR.',
    author='david.sulc',
    author_email='sulda.str@gmail.com',
    # packages=['tacrpy', 'nlp', 'datahub']
    packages=find_packages(),
    project_urls={
        'Documentation': 'https://data.tacr.cz/tacrpy/docs/index.html'
    },
    install_requires=[
        'pandas',
        'gspread',
        'numpy',
        'sentence-transformers',
        'nltk',
        'requests',
        ]
)



"""Soubor funkcí k práci s OpenAPI DataHubu."""

import requests


def post_data(json_data: list[dict], test: bool = True) -> int:
    """ Nahraje přes OpenAPI data do DataHubu.

    Než se data nahrají je vhodné provést kontroly správnosti dat.

    :param json_data: data ve formátu JSON, které se mají nahrát do DataHubu
    :param test: nahraje data buď na testovací nebo produkční prostředí
    :return: status code
    """

    if test:
        server = 'http://datahub-test.tacr.cz:8080/'
    else:
        server = 'http://datahub.tacr.cz:8080/'

    query = 'openapi/entities/v1/'
    url = server + query

    r = requests.post(url, json=json_data)

    return r.status_code


def get_data(urn: str, test: bool = True) -> dict:
    """ Získá data k vybrané entitě přes OpenAPI.

    :param urn: URN entity, pro kterou chceme získat data
    :param test: získá data buď z testovacího nebo produkčního prostředí
    :return: data ve formátu JSON
    """

    if test:
        server = 'http://datahub-test.tacr.cz:8080/'
    else:
        server = 'http://datahub.tacr.cz:8080/'

    query = f'openapi/entities/v1/latest?urns={urn}'

    url = server + query

    r = requests.get(url)

    json_data = r.json()['responses']

    return json_data


def delete_data(urn: str, test: bool = True, soft: bool = False) -> int:
    """ Vymaže data vybrané entity přes OpenAPI

    :param urn: URN entity, kterou chceme vymazat
    :param test: vymaže data z testovacího nebo produkčního prostředí
    :param soft: můžeme entitu vymazat pouze z vyhledávání (soft = True) nebo kompletně (soft = False)
    :return: status code
    """

    if test:
        server = 'http://datahub-test.tacr.cz:8080/'
    else:
        server = 'http://datahub.tacr.cz:8080/'

    if soft:
        soft_delete = '&soft=true'
    else:
        soft_delete = '&soft=false'

    query = f'openapi/entities/v1/?urns={urn}'

    url = server + query + soft_delete

    r = requests.delete(url)

    return r.status_code
